%% Fd support for Curry.
%%
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module('Fdlib', []).

:- use_module('Core', [tonf/2, tohnf/2]).

:- multifile     user_tohnf/2, user_tonf/2.
:- discontiguous user_tohnf/2, user_tonf/2.

%% We are going to support fd.
:- use_package(fd).

%% The support is incomplete, and Ciao version is evolving yet.
user_tohnf('Fd_in_range'(A, 'Prelude_,'(R1, R2))#A1, R) :-
	tonf(A, A1),
	A1 in R1..R2,
	R = 'Prelude_Success'.

user_tonf(A + B, R) :-
	tonf(A, A1),
	tonf(B, B1),
	R = A1 + B1.

user_tonf(A - B, R) :-
	tonf(A, A1),
	tonf(B, B1),
	R = A1 - B1.

user_tohnf('Fd_.+.'(A, B)#R, R) :-
	R = (A + B).

user_tohnf('Fd_.-.'(A, B)#R, R) :-
	R = (A - B).

user_tohnf('Fd_.<.'(A, B)#R, R) :-
	tonf(A, A1),
	tonf(B, B1),
	A1 .<. B1,
	R = 'Prelude_Success'.

user_tohnf('Fd_.=<.'(A, B)#R, R) :-
	tonf(A, A1),
	tonf(B, B1),
	A1 .=<. B1,
	R = 'Prelude_Success'.

user_tohnf('Fd_.>.'(A, B)#R, R) :-
	tonf(A, A1),
	tonf(B, B1),
	A1 .>. B1,
	R = 'Prelude_Success'.

user_tohnf('Fd_.>=.'(A, B)#R, R) :-
	tonf(A, A1),
	tonf(B, B1),
	A1 .>=. B1,
	R = 'Prelude_Success'.

user_tohnf('Fd_.=.'(A, B)#R, R) :-
	tonf(A, A1),
	tonf(B, B1),
	A1 .=. B1,
	R = 'Prelude_Success'.

user_tohnf('Fd_.<>.'(A, B)#R, R) :-
	tonf(A, A1),
	tonf(B, B1),
	A1 .<>. B1,
	R = 'Prelude_Success'.

%% Labelling predicates.
user_tohnf('Fd_all_different'(List)#R, R) :-
	tonf(List, List1),
	all_different(List1),
	R = List1.

user_tohnf('Fd_labeling'(List)#R, R) :-
	tonf(List, List1),
	labeling(List1),
	R = List1.
