%% Monadic IO support for Curry.
%%
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module('IOlib', []).

:- use_module('Core', [tonf/2, tohnf/2]).

:- multifile     user_tohnf/2, user_tonf/2.
:- discontiguous user_tohnf/2, user_tonf/2.

:- use_module(library(lists),         [list_concat/2]).
:- use_module(library(strings),       [write_string/1]).
:- use_module(library(format),        [format_to_string/3]).
:- use_module(library(iso_byte_char), [get_char/1]).
:- use_module(library(terms),         [atom_concat/2] ).

:- use_module(library(hiordlib)).

%% IO ops.
user_tohnf('IO_putStr'(S)#B, B) :- 
	!,
	tonf(S, St),
	write_string(St),
	B = 'IO_World'('Prelude_()').

user_tohnf('IO_putStrLn'(S)#B, B) :- 
	!,
	tonf(S, St),
	write_string(St), nl,
	B = 'IO_World'('Prelude_()').

user_tohnf('IO_getChar'#B, B) :-
	!,
	get_code(C),
	B = 'IO_World'(C).

user_tohnf('IO_show'(A)#B, B) :-
	(
	    tonf(A, C),!
	;
	    A = C
	),
	sloth_show(C, B).
 
%% hackish show until we have type classes.
sloth_show(A, B) :-
	pretty_prolog(A, AP),
	!,
	format_to_string("~p", [AP], B).

pretty_prolog(A, B) :-
	var(A),
	!,
	A = B.

pretty_prolog(A, B) :-
	number(A),
	!, 
	A = B.

pretty_prolog('Prelude_,'(A, B), (A1,B1)) :-
	pretty_prolog(A, A1),
	pretty_prolog(B, B1).

pretty_prolog(A, B) :-
	functor(A, _, 0),
	split_id(A, B),
	!.

pretty_prolog(A, B) :-
	A =.. [Name|Args],
	split_id(Name, IName),
	!,
	map(Args, pretty_prolog, IArgs),
	B =.. [IName|IArgs].

split_id(Id, SId) :-
	atom_codes(Id, Sr),
	append(Mod, [0'_|Name], Sr), 
	atom_codes(ModA, Mod),
	atom_codes(NameA, Name),
	atom_concat([ModA, '.', NameA], SId),
	!.

split_id(Id, Id).

:- multifile portray/1.
portray((A,B)) :- 
	write('('), 
	write(A), 
	write(','), 
	write(B),
	write(')').


