%% Sloth's main module.
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(main, [main/0]).

:- use_module(library(lists),   [append/3]).
:- use_module(library(system),  [file_exists/1]).
:- use_module(library(write),   [write/1]).
:- use_module(library(terms),   [atom_concat/2]).
:- use_module(library(filenames),  [file_name_extension/3]).

:- use_module(error,        [do_error/1]     ).
:- use_module(comp_manager, [compileFile/2]  ).
:- use_module(cmdopts,      [process_options/0, cmd_options/2, help/0] ).
:- use_module(shell,        [start_shell/0]).

% egallego@babel: Sloth main module.

main :-
	catch(sloth_main, E, do_error(E)).

sloth_main :-
	(
	    process_options,
	    cmd_options('File', File) ->
	    check_file(File),
	    make_pl_suffix(File, FileOut),
	    !,
	    compileFile(File, FileOut),
	    write('File '), write(File), 
	    write(' is compiled to '), write(FileOut), 
	    write('\n')
	;
	    cmd_options('-i', _),
	    start_shell
	).

sloth_main :-
	help.

% Checks the file exists.
check_file(File):-
	(
	    file_exists(File) -> 
	    true
	;    
	    throw(error(main, no_such_file(File)))
        ).

% Changes a f.curry to f.pl
make_pl_suffix(OldFile, NewFile) :-
	file_name_extension(OldFile, File, '.curry'),
	file_name_extension(NewFile, File, '.pl').

make_pl_suffix(_, _) :- 
	!,
	throw(error(main, 'Error, filename must end in .curry')).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

