%% Curry to prolog main translator.
%% Copyright (C) 1998-2006, Babel Group
%%
%% Written by: 
%%   * Julio Mariño Carballo
%%   * Jose María Rey
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(trans_lazy, [dump/1, dump_module/2, tr_expr/2], [assertions]).

:- use_module(deftree,       [defTree/3, call_defTree/2] ).
:- use_module(symtab,        [get_rules/2, get_real_rules/2, 
                              get_object_property/3, object_arity/2,
			      get_module_functions/2] ).
:- use_module(trans_cons,    [translate_constructors/1] ).
:- use_module(curryast,      [make_name/2, get_function/2] ).
:- use_module(const_expr,    [is_constant/1]).
:- use_module(cmdopts,       [cmd_options/2] ).
:- use_module(typer,         [infer_type_expr/2]).
%
:- use_module(library(lists),      [append/3, length/2, reverse/2, 
                                    list_concat/2, contains_ro/2] ).
:- use_module(library(aggregates), [findall/3] ).
:- use_module(library(write),      [write/1, write_term/2, portray_clause/1] ).
:- use_module(library(terms),      [atom_concat/2]).

:- include(ast_ops).
:- include(type_ops). %% Arggg.

%% subst_term(-Term, -Pos, -Exp, +NExp)
%% Pos is a list of positions.
subst_term(_Term,[],Subterm,Subterm) :-
	!.

subst_term(Term, Pos, Subterm, Term2) :-
	functor(Term, Functor, Arity),      
		%Cambia el termino en la posicion indicada
	functor(Term2, Functor, Arity),     % y lo sustituye por Subterm.
	subst_term_aux(Term, Term2, Pos, Subterm, Arity).

% FIXME: Document this.
subst_term_aux(_, _, _, _, 0) :-
	!.
subst_term_aux(Term, Term2, [N|Ps], Subterm, N) :-
	!,
	arg(N,Term,T),
	subst_term(T,Ps,Subterm,Term3),
	arg(N,Term2,Term3),
	N1 is N-1,
	subst_term_aux(Term,Term2,[N|Ps],Subterm,N1).

subst_term_aux(Term,Term2,[N|Ps],Subterm,N2) :-
	\+ ( N = N2 ),
	arg(N2,Term,T),
	arg(N2,Term2,T),
	N1 is N2-1,
	subst_term_aux(Term,Term2,[N|Ps],Subterm,N1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

parser2trad(freevar(X), X) :-
	!.

parser2trad(param(X), X) :-
 	!.

parser2trad(const(_, X), X) :- 
	!.

parser2trad(qid(M, Id), P1) :-
	make_name(qid(M, Id), P1),
	!.

parser2trad(P@Ts,T) :-
	!,
	P = qid(_, _),
	make_name(P, P1),
	parser2trad_list(Ts,Ts2),
	T =.. [P1|Ts2].

parser2trad_list([], []).
parser2trad_list([T|Ts], [T2|Ts2]) :-
	parser2trad(T,T2),
	parser2trad_list(Ts,Ts2).

%% trans_struc_goal(-ParserExp, +PrologExp)
%% ParserExp is in a@[Args] form, while PrologExp is in
%% final prolog form.

trans_struc_goal(param(V)@[], V) :- !.
trans_struc_goal(param(V)@Pars, memo(apply_list(V, Pars2))) :-
	!,
	trans_struc_goal_list(Pars,Pars2).

%% This kind of application flattening has to improve.
trans_struc_goal((F@Arg1)@Arg2, Res) :-
	append(Arg1, Arg2, ArgN),
	!,
	trans_struc_goal(F@ArgN, Res).

%% Translation of partial applications (both for functions and cons)
trans_struc_goal(FId@Pars, pap(MI, Quedan, Pars2)) :-
	get_rules(FId, _Rules), % Is a function
	object_arity(FId, Arity),
	length(Pars,NumPars),
	NumPars < Arity,
	!,
	Quedan is Arity - NumPars,
	reverse(Pars, ParsRev),
	make_name(FId, MI),
	trans_struc_goal_list(ParsRev,Pars2).

trans_struc_goal(FId@Pars, pap(XX, Quedan,Pars2)):- %% Constr. currif.
	get_object_property(FId, constructs, _Type),
	make_name(FId, XX),
	object_arity(FId, Arity),
	length(Pars, NumPars),
	NumPars < Arity,
	!,
	Quedan is Arity - NumPars,
	reverse(Pars, ParsRev),
	trans_struc_goal_list(ParsRev,Pars2).

%% Next 3 are for single cons or functions.
trans_struc_goal(FId, X) :-
	get_rules(FId, _Rules),
	!,
	make_name(FId, XX),
	object_arity(FId, Arity),
	(
	    Arity == 0 -> 
	    X = memo(XX)
	; 
	    X = pap(XX, Arity, [])
 	).

trans_struc_goal(FId@[], X) :-
	!,
	trans_struc_goal(FId, X).

trans_struc_goal(FId, X) :-
	get_object_property(FId, constructs, _Type),
	make_name(FId, XX),
	object_arity(FId, Arity),
	(
	    Arity == 0 -> 
	    X = XX
	; 
	    X = pap(XX, Arity, [])
 	).

%% Generic translation. Both for cons and for fun.
trans_struc_goal(FId@Pars, T) :- 
	get_object_property(FId, constructs, _T),
	!,
	trans_struc_goal_list(Pars, Parst),
	make_name(FId, XX),
	T =.. [XX|Parst].

trans_struc_goal(FId@Pars, memo(T)) :- 
	object_arity(FId, Arity),
	length(Pars, Arity),
	!,
	trans_struc_goal_list(Pars, Parst),
	make_name(FId, XX),
	T =.. [XX|Parst].

%% This simulates an over-app.
trans_struc_goal(FId@Pars, T) :-
	object_arity(FId, ArityF),
	length(Pars, ArityP),
	ArityP > ArityF,
	!,
	list_take(Pars, ArityF, ParsC),
	trans_struc_goal(FId@ParsC, CompleteApp),
	append(ParsC, ParsR, Pars),
	trans_struc_goal_list(ParsR, ExtraArgs),
	T = memo(apply_list(CompleteApp, ExtraArgs)).

%% This is the most awkward rule, for translating cond exps we have to
%% look up the type, and call a different function depending on it.
%% Fixme, this should be cleanly implemented by a rewriting.
trans_struc_goal(untyped_cond(List), R) :-
	copy_term(List, List1),
	infer_type_expr(List1, Type),
	Type = tid('Prelude', [])<@[tid('Prelude', tuple)<@[CondT, _B]],
	!,
	(
	    CondT = tid('Prelude', 'Bool')<@[] ->
	    Expr = qid('Prelude', bool_cond)@[List],
	    trans_struc_goal(Expr, R)
	;
	    CondT = tid('Prelude', 'Success')<@[] ->
	    Expr = qid('Prelude', success_cond)@[List],
	    trans_struc_goal(Expr, R)
	;
	    throw(error(translator, not_allowed_type_for_cond(CondT)))
	).

%% All the simple cases
trans_struc_goal(E, E2):-
	parser2trad(E,E2). 
  
trans_struc_goal_list([],[]).
trans_struc_goal_list([A|As], [B|Bs]) :-
	trans_struc_goal(A,B),
	trans_struc_goal_list(As,Bs).

:- pred list_take(L, N, LL) : list * int * list # "Takes N elements from L.".
list_take(_L, 0, []).
list_take([L|LR], N, [L|RR]) :-
	N1 is N - 1,
	list_take(LR, N1, RR).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Memoization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

memberM(X, [Y|V]) :-
	var(V),
	var(Y),
	X = Y,
	!.

memberM(X#Res, [Y#Res|_]) :-
	X == Y.
memberM(X, [_|XS]) :-
	memberM(X, XS).


memoize(V, V, _Dicc) :-
	var(V).

memoize(memo(Exp), Exp1#Res, Dicc) :-
	!,
	memoize(Exp, Exp1, Dicc),
	memberM(Exp1#Res, Dicc).	

memoize(CExp, NExp, Dicc) :-
	CExp =.. [F|Args],
	memoize_list(Args, NArgs, Dicc),
	NExp =.. [F|NArgs].

memoize_list([], [], _Dicc).
memoize_list([R|RS], [S|SS], Dicc) :-
	memoize(R, S, Dicc),
	!,
	memoize_list(RS, SS, Dicc).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Aux predicates.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Create fresh vars. Must see if CIAO has a builtin pred for doing this.
%% We can use length(Res, N), it think.

new_vars(0, []) :- !.
new_vars(N, [_|Ts]) :-
	N1 is N-1,
	new_vars(N1, Ts).

%% This is like args but structurally recursive
term_at(Term,[],Term).
term_at(Term,[Pos|Poss],Subterm):-
	arg(Pos,Term,Term2),
	term_at(Term2,Poss,Subterm).

%% Generates a [[3],[2],[1]] list.
comienzos(0, []) :- !.
comienzos(N, [[N]|R]) :-
	N > 0,
	N1 is N-1,
	comienzos(N1, R).

%% Suffix a number or letter to the end of Pred.
letra_nivel(Pred, 0, Pred).
letra_nivel(Pred, Nivel, PredNivel):-
	name(Pred, PredList),
	Letra is 0'a + Nivel - 1,
	append(PredList, [Letra], PredList2),
	name(PredNivel, PredList2).

numero_nivel(Pred, Nivel, PredNivel) :-
	name(Pred, PredList),
	Letra is 0'0 + Nivel ,
	append(PredList, [Letra], PredList2),
	name(PredNivel, PredList2).
	
%% Get the name of the deftree translation for a predicate.
%% deftree_pred_name(-Pred, -Level, +NewPred)
deftree_pred_name(Pred, Level, NewPred) :-
	name(Pred, PredStr),
	(
	    Level = 0 ->
	    Sufijo = [0'1]
        ;
	    Letra is 0'a + Level - 1,
	    Sufijo = [0'1, Letra]
        ),
	append(PredStr, Sufijo, PredStr1),
	name(NewPred, PredStr1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% Part 1, tohnf translation.

tohnf_entries(Module, Reglas):-
	get_module_functions(Module, Functions),
	map_tohnf_entry(Functions, Reglas). 

%% Calls to_hnf_entry for each F in the list.
map_tohnf_entry([],[]).
map_tohnf_entry([F|Fs],Rs):-
	call_defTree(F, Tree),
	tohnf_entry(F,Tree, 0, R),
	Rs3 = [],
	append(R, Rs3, Rs3b),
	map_tohnf_entry(Fs, Rs2),
	append(Rs3b,Rs2,Rs).

%%% tohnf_entry(-Pred, -Deftree, -NumLetra, +Rule)
%%% Rule is a list of prolog clauses.
%%%
%%% This pred calls the deftree for a function, assuring
%%% that the branch arg is in hnf.

tohnf_entry(Pred, Deftree, NumLetra, Rule) :- 
	Head = user_tohnf(TermPred#Res, Res),  
	object_arity(Pred, Arity),
	new_vars(Arity, Vars),
	make_name(Pred, NPred),
	TermPred =.. [NPred|Vars],
	(
	    Deftree = branch(Pos, _)-_ ->
	    %% Get the argument that deftree is going to
	    %% switch on and tohnf it.
	    Rule = [(Head :- Body)],
	    term_at(TermPred, Pos, Var),
	    Body = (
		       tohnf(Var, Free1), 
		       TermResult
		   ),
            append(Vars, [Res], Vars2),
	    deftree_pred_name(NPred, NumLetra, PredDefTree),
	    TermResult_aux =.. [PredDefTree | Vars2],
	    subst_term(TermResult_aux, Pos, Free1, TermResult)
	;
	    Deftree = regla-_ ->
	    deftree_pred_name(NPred, NumLetra, PredDefTree),
	    append(Vars, [Res], Vars2),
	    BodyR =.. [PredDefTree | Vars2],
	    Rule  = [(Head :- BodyR)]
	;
	    Deftree = or(Deftrees)-_ ->
	    map_or_trees(Pred, Deftrees, 0, Rule) 
	;
	    Rule = [(Head :- fail)]
	).

%% map_or_trees(Pred,Arboles,Num_Rama,Reglas): Aplica tohnf_entry a
%% cada una de las alternativas de un arbol or y devuelve el resultado
%% en reglas (una lista de ellas).
map_or_trees(_, [], _, []).
map_or_trees(Pred, [T|Ts], N, Z) :-
	N1 is N+1,
	tohnf_entry(Pred, T, N, R),
	map_or_trees(Pred, Ts, N1, Rs),
	append(R, Rs, Z).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Part 2, deftree translation.

trans_deftrees(Module, Reglas):-
	get_module_functions(Module, Functions),
	map_trans_deftree(Functions, Rs), 
	list_concat(Rs, Reglas).

%% Can trans_deftree fail?
map_trans_deftree([],[]).
map_trans_deftree([F|Fs],[Lr|Lrs]):-
	(
	    trans_deftree(F, Lr) ->
	    true
	;
	    throw(error(translator, 'Cannot generate a definitional tree'(F)))
	),
	!,
	map_trans_deftree(Fs,Lrs).


%% This predicate translates the definitional tree of a predicate.
%% It has three cases: branch, or and rule.

trans_deftree(Pred, Reglas):-
	call_defTree(Pred, Deftree),
	Deftree = or(Arboles)-_,
	!,
	object_arity(Pred, Arity),
	new_vars(Arity, Vars),
	make_name(Pred, NPred),
	T_Acum =.. [NPred|Vars],
	map_trans_deftree_aux(Pred, 1, T_Acum, Arboles, 0, ReglasAux),
	list_concat(ReglasAux, Reglas).

trans_deftree(Pred, Reglas):-
	object_arity(Pred, Arity),
	get_rules(Pred, ReglasPred),
	comienzos(Arity, Comienzos),
	new_vars(Arity, Vars),
	make_name(Pred, NPred),
	TermAcum =.. [NPred|Vars],
	defTree(ReglasPred, Comienzos, DTree),
	!,
	(
	    DTree = branch(Pos,Arbol)-_Reglas ->
	    trans_deftree_aux(Pred, 1, TermAcum, (Pos,Arbol), 0, Reglas) 
	;
	    DTree = regla-[N] ->
	    trans_deftree_aux(Pred, 1, TermAcum, regla-[N], 0, Reglas)
	).  

%%% trans_deftree_aux(-Predicado, -Nivel, -Termino_acum, -Arbol, -NumArbol, +Reglas).
%%%
%%% The only thing this predicate does is to generate the number and the 
%%% name of predicates required for the tranlation, performing the actual 
%%% body translation by tr_rule.

trans_deftree_aux(_Predicado, _Nivel, _T_Acum, fallo-[], _, []) :- !.

trans_deftree_aux(Pred, Nivel, T_Acum, (or(Arboles)-_), _NumArbol, Reglas) :-
	!,
	map_trans_deftree_aux(Pred, Nivel, T_Acum,Arboles, 0, Reglas2),
	list_concat(Reglas2, Reglas).
  
trans_deftree_aux(Pred, Nivel, _T_Acum, regla-[Reg], NumArbol, [(Head :- Body)]) :-
	!,
	make_name(Pred, NPred),
	numero_nivel(NPred, Nivel, PredNumber),
	tr_rule(Reg, (Head1 :- Body)),
	Head1 =.. [_ | Args],
	letra_nivel(PredNumber, NumArbol, PredNb),
	Head =.. [PredNb | Args].

trans_deftree_aux(Pred, Nivel, T_Acum, branch(Pos, Ramas)-_, NumArbol, [(Head :- Body) | Rs]) :-
	make_name(Pred, NPred),
	N1 is Nivel + 1,
	numero_nivel(NPred,  Nivel,    PredN),
	letra_nivel( PredN,  NumArbol, PredNb),	
	numero_nivel(NPred,  N1,       PredN1),	
	letra_nivel( PredN1, NumArbol, PredN1b),
	
	term_at(T_Acum, Pos, T1),
	
        T_Acum =.. [_|T_AcumList],
	append(T_AcumList, [Free], HeadPars),
	
	Head =.. [PredNb | HeadPars], 

	subst_term(T_Acum, Pos, Free2, T_Acum2),
	T_Acum2 =.. [_|T_Acum2List],
	append(T_Acum2List, [Free], Head2Pars),
	letra_nivel(PredN1, NumArbol, PredN1b),
	TermPred1 =.. [PredN1b|Head2Pars],	
	Body = (
		   tohnf(T1, Free2),
		   TermPred1
	       ),
	N1 is Nivel + 1,
	trans_deftree_aux(Pred, N1, T_Acum, (Pos, Ramas), NumArbol, Rs),
	!.

trans_deftree_aux(Pred, Nivel, T_Acum, (Pos,[(Cons,Tree)|Ts]), NumArbol, Rs) :- !,
	(
	    is_constant(Cons) ->
	    ArityCons = 0 
	; 
	    object_arity(Cons, ArityCons)
	), 
	make_name(Cons, NCons),
	new_vars(ArityCons, Vars),
	TermCons =.. [NCons|Vars],
	subst_term(T_Acum, Pos, TermCons, T_Acum2),
	trans_deftree_aux(Pred, Nivel, T_Acum2, Tree, NumArbol, Rs1),
	trans_deftree_aux(Pred, Nivel, T_Acum, (Pos, Ts), NumArbol, Rs2),
	append(Rs1, Rs2, Rs3),
	Rs = Rs3.

%% Should output error or fail.
trans_deftree_aux(_,_,_,(_,[]),_,[]).

%% Only used for or- type deftrees.
map_trans_deftree_aux(_,_,_,[],_,[]).
map_trans_deftree_aux(Pred,Nivel,T_Acum,[Tree|Trees],NumArbol,Reglas):-
	NumArbol1 is NumArbol + 1,
	trans_deftree_aux(Pred, Nivel, T_Acum, Tree, NumArbol, ReglasTree),
	map_trans_deftree_aux(Pred, Nivel, T_Acum, Trees, NumArbol1, ReglasTrees),
	append([ReglasTree], ReglasTrees, Reglas).

%%%%%%%%%%%
% Rule translation.

%% new possible helper preds, change_head and change_body.

tr_expr(Ast, Prolog) :-
	trans_struc_goal(Ast, TExpr),
	memoize(TExpr, Prolog, _Dicc).

% Caso 1, la regla empieza con un constructor.

%% We generate tohnf(pred(P1..Pn), 'Cons(Q1...Qn)') 
tr_rule(R, Rt) :-
	R = (Head:=Body),
	(
	    ( Body = qid(M, C)@_ )
	;
	    ( Body = qid(M, C) ) 
	),
	get_object_property(qid(M,C), constructs, _Type),
	!,
	tr_expr(Body, BodyTrad),
	parser2trad(Head, HeadTrad),
	HeadTrad =.. [Pred|Hlist],
	append(Hlist, [BodyTrad], Hlist2),
	Ht =.. [Pred|Hlist2],
	Rt = (Ht :- true).

tr_rule(R,Rt) :-
	R = (Head := Body),
	get_function(Head, QPred),
	parser2trad(Head, HeadTrad),
	HeadTrad =.. [Pred|Hlist],
	append(Hlist, [R0], Hlist2),
	HeadTrad2 =.. [Pred|Hlist2],
	tr_expr(Body, BodyTrad),
	BodyTrad2 = (
			tohnf(BodyTrad, R0)
		    ),
	% Este es el en el cual la cabecera de la función toma menos argumentos
	% que los necesarios -> debemos añadir más variables.
	Rt_intermedio = (HeadTrad2 :- BodyTrad2),  
	(
	    get_rules(QPred, _Rules),
	    object_arity(QPred, Type_arity),
	    length(Hlist, Prog_arity),
	    %% Reglas con aridad inferior a la del tipo (prueba.bab)
	    Type_arity > Prog_arity ->                    
	    Dif is Type_arity - Prog_arity,
	    new_vars(Dif, Vars),
	    %%  reverse(Vars, VarsRev),
            append(Hlist, Vars, Hlist3),
	    append(Hlist3, [R0], Hlist4), %Parametro de salida
	    Head3 =.. [Pred|Hlist4],
            Body3 = (
			tohnf(apply_list(BodyTrad, Vars)#R0, R0)
		    ),
	    Rt = (Head3 :- Body3)
	;
	    Rt = Rt_intermedio
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Module support.

write_module_header(File, Module) :-
	get_object_property(Module, exports, Exports),
	(
	    get_rules(qid(Module, main), _) ->
	    ModuleExports = [exports/1, main/0]
	;
	    ModuleExports = [exports/1]
	),
	portray_clause((:- module(File, ModuleExports))),
	nl,
	static_header,
	nl,
	write_search_path,
	nl,
	write_core,
	nl,
	findall(Imp, get_object_property(Module, imports, Imp), Imports),
	write_imports(Imports),
	nl,
	write_exports(Exports).

write_core :-
	cmd_options('-l', [LibDir]),
	atom_concat([LibDir, '/', 'Core'], CorePath),
	portray_clause((:- use_module(CorePath, [tonf/2, tohnf/2]))).

%%
write_search_path :-
	cmd_options('-l', [LibDir]),
	portray_clause(file_search_path(sloth, LibDir)),
	(
	    LibDir = '.' -> % Include . in the search path for normal files.
	    true
	;
	    portray_clause(file_search_path(sloth, '.'))
	).

%%
write_exports([]) :- !.
write_exports([id(F)|Exp]) :-
	get_object_property(F, type, Type),
	( 
	    get_object_property(F, assoc, Assoc),
	    get_object_property(F, priority, Prio) ->
	    OpAttributes = op(Assoc, Prio)
	;
	    OpAttributes = none
	),
	portray_clause(exports(id(F, Type, OpAttributes))),
	write_exports(Exp).

write_exports([type(Name, Cons)|Exp]) :-
	export_cons(Cons, ConsExp),
 	portray_clause(exports(type(Name, ConsExp))),
 	write_exports(Exp).

export_cons([], []).
export_cons([C|CN], [(C, CType)|CEN]) :-
	get_object_property(C, type, CType),
	export_cons(CN, CEN).

%%
write_imports(L) :-
       clause_write_imports(L, Imports),
       portray_clause((:- initialization(Imports))).

%%
clause_write_imports([], (!)).
clause_write_imports([M|Rest], (use_module(sloth(M)), Imports) ) :-
       clause_write_imports(Rest, Imports).

%%
static_header :-
	portray_clause((:- use_module(library(system)))),
	portray_clause((:- use_module(library(lists)))),
	nl,
	portray_clause((:- dynamic       file_search_path/2)),
	portray_clause((:- multifile     is_constructor/1, user_tonf/2, user_tohnf/2, file_search_path/2)),
	portray_clause((:- discontiguous is_constructor/1, user_tonf/2, user_tohnf/2)),
	nl,
	%% For types export, umm this should done automatically??
	portray_clause((:- op(651, xfy, <@))),  % Typevar aplication.
	portray_clause((:- op(652, xfy, -->))), % Normal apply operator.
	portray_clause((:- op(653, xfy, =>))).  % Type class constraint.

%% Main translation function.
dump(Module) :-
	write_module_header(Module, Module),
	!,
	translate_constructors(ConsP), write_prolog(ConsP),
	!,
	tohnf_entries(Module, TohnfR), write_prolog(TohnfR),
	nl, !,
	trans_deftrees(Module, DefTreeR), write_prolog(DefTreeR),
	nl, !,
	write_main_goal.

dump_module(File, Module) :-
	write_module_header(File, Module),
	!,
	translate_constructors(ConsP), write_prolog(ConsP),
	!,
	tohnf_entries(Module, TohnfR), write_prolog(TohnfR),
	nl, !,
	trans_deftrees(Module, DefTreeR), write_prolog(DefTreeR),
	nl, !,
	write_main_goal.


%% Temp:
write_prolog([]).
write_prolog([C|CL]) :-
	portray_clause(C), 
	nl,
	write_prolog(CL).

%% If we have a main function, link it with the prolog predicate main.
%% Fixme: Type checking doesn't belong here.
write_main_goal :-
	get_rules(qid(Module, main), _),
	!,
	get_object_property(qid(Module, main), type, Type),
	(
	    Type = tid('IO', 'IO')<@[tid('Prelude', void)<@[]] ->
	    make_name(qid(Module, main), Main),
	    portray_clause(( main :- tohnf(Main#R, R),
                                 tonf(R, _Rnf)))
	;
	    throw(error(compiler, main_has_to_be_IO_but_is(Type)))
	).

write_main_goal.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:
