%% Sloth's identifiers resolution.
%%
%% Copyright (C) 2005-2006, Babel Group
%%
%% Written by: 
%%   * Emilio Jesús Gallego Arias
%%
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

:- module(unref, [solve_all_refs/0, solve_refs/2, 
	          free_all_vars/0, free_vars/3,
	          solve_all_type_refs/0, solve_type_refs/2, 
		  free_all_type_vars/0, free_type_vars/3]).

:- use_module(scope,    [lookup_symbol/2]).
:- use_module(curryast, [map_rules/1]).
:- use_module(types,    [map_types/1]).

%% TODO, at least free_vars/0 is obsolete, parser should call free_rule when 
%% appropriate. Maybe resolve_refs is obsolete too.
solve_all_refs :-
	map_rules(solve_refs).

%% Rename param(x) to param(X) in all the expressions.
free_all_vars :- 
	map_rules(free_vars(_Dicc)).

solve_all_type_refs :-
	map_types(solve_type_refs).

free_all_type_vars :-
	map_types(free_type_vars(_Dicc)).

%% solve_refs: try to replace unref/1 terms, by their definition.
%% If the symbol is undefined, an exception is thrown.

solve_refs(unref(X), E) :-
	(
	    lookup_symbol(X, E)
	;
	    throw(error(symbol_resolution, symbol_undefined(X)))
	).

%% free_expr: changes all the fixed variables in a rule by a Prolog
%% free var, assigning the same Var to the same identifiers.
free_vars(param(X), _Dicc, param(_)) :-
	\+ var(X),
	X = '_',
	!.

free_vars(param(X), Dicc, param(V)) :-
	\+ var(X),
	!,
	member(X-V, Dicc).

free_vars(freevar(X), Dicc, freevar(V)) :-
	\+ var(X),
	!,
	member(X-V, Dicc).

%% Same for types.
solve_type_refs(T, E) :-
	\+ var(T),
	unref(X) = T,
	(
	    lookup_symbol(X, E)
	;
	    throw(type_undefined(X))
	).

%% Some care because the typer guys want free vars as types.
free_type_vars(T, Dicc, V) :-
	\+ var(T),
	tvar(X) = T,
	\+ var(X),
	!,
	member(X-V, Dicc).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

